import numpy
import pandas

from . import _slimp

def r_squared(*args, **kwargs):
    # https://avehtari.github.io/bayes_R2/bayes_R2.html
    
    from .model import Model
    
    if len(args) == 1 and isinstance(args[0], Model):
        return _r_squared_model(*args, **kwargs)
    elif len(args) == 2 and isinstance(args[0], pandas.DataFrame):
        return _r_squared_data_frame(*args, **kwargs)
    else:
        raise NotImplementedError()

def _r_squared_model(model):
    if isinstance(model.formula, list) and isinstance(model.formula[1], tuple):
        return r_squared(model.posterior_epred, model.draws["sigma_y"])
    elif isinstance(model.formula, list):
        df = pandas.concat(
            [
                r_squared(
                    model.posterior_epred.filter(regex=fr"mu\.[^.]+\.{i+1}"),
                    model.draws[f"{c}/sigma"])
                for i, c in enumerate(model.outcomes.columns)],
            axis="columns")
        df.columns = model.outcomes.columns
        return df
    else:
        return r_squared(model.posterior_epred, model.draws["sigma"])

def _r_squared_data_frame(mu, sigma):
    var_mu = mu.var("columns")
    var_sigma = sigma**2
    return var_mu/(var_mu+var_sigma)

def hmc_diagnostics(data, max_depth):
    energy = data.sel(parameter="energy__").values
    diagnostics = pandas.DataFrame({
        "divergent": data.sel(parameter="divergent__").sum(axis=1),
        "depth_exceeded": 
            (data.sel(parameter="treedepth__")>=max_depth)
            .sum(axis=1),
        "e_bfmi": 
            numpy.square(numpy.diff(energy, axis=1)).mean(axis=1) 
                / numpy.var(energy, axis=1, ddof=1)})
    return diagnostics

def summary(data, percentiles=(5, 50, 95)):
    summary = {}
    
    summary["Mean"] = numpy.mean(data, axis=(1,2))
    summary["MCSE"] = None
    summary["StdDev"] = numpy.std(data, axis=(1,2))
    quantiles = numpy.quantile(data, numpy.array(percentiles)/100, axis=(1,2))
    for p, q in zip(percentiles, quantiles):
        summary[f"{p}%"] = q
    
    summary["N_Eff"] = _slimp.get_effective_sample_size(data)
    summary["R_hat"] = _slimp.get_split_potential_scale_reduction(data)
    
    summary["MCSE"] = numpy.sqrt(summary["StdDev"])/numpy.sqrt(summary["N_Eff"])
    
    return pandas.DataFrame(summary, index=data["parameter"])

def hdi(x, mass):
    """ Highest density interval, after "Doing Bayesian Data Analysis",
        J. Kruschke, section 25.2.3
    """
    
    x = numpy.sort(x)
    N = len(x)
    # Number of points in the interval to account for the mass
    count = int(numpy.ceil(mass * N))
    # Widths of all intervals of given mass
    widths = x[count:] - x[:N - count]
    # HDI is the narrowest interval
    min_index = numpy.argmin(widths)
    return (x[min_index], x[min_index+count])



from ._slimp import action_parameters
from .misc import sample_data_as_df, sample_data_as_xarray
from .model import Model
from .plots import KDEPlot, parameters_plot, predictive_plot
from .samples import Samples
from .stats import hmc_diagnostics, r_squared, summary

from . import multilevel, multivariate, univariate



import formulaic
import numpy
import pandas

from .predictor_mapper import PredictorMapper

class ModelData:
    def __init__(self, formula, data):
        self.formula = formula
        self.data = data
        
        self.outcomes, self.predictors = formulaic.model_matrix(formula, data)
        
        self.predictor_mapper = PredictorMapper(self.predictors, self.outcomes)
        
        mu_y = numpy.mean(self.outcomes.values)
        sigma_y = numpy.std(self.outcomes.values)
        sigma_X = numpy.std(
            self.predictors.filter(regex="^(?!.*Intercept)").values, axis=0)
        
        self.fit_data = {
            "N": len(data), "K": self.predictors.shape[1],
            "y": numpy.squeeze(self.outcomes), "X": self.predictors,
            
            "mu_alpha": mu_y, "sigma_alpha": 2.5*sigma_y,
            "sigma_beta": 2.5*sigma_y/sigma_X,
            "lambda_sigma": numpy.squeeze(1/sigma_y)}
    
    def new_predictors(self, data):
        data = data.astype({
            k: v for k, v in self.data.dtypes.items() if k in data.columns})
        predictors = pandas.DataFrame(
            formulaic.model_matrix(self.formula.split("~")[1], data))
        return predictors
    
    def new_data(self, X_new=None):
        if X_new is None:
            X_new = self.fit_data["X"]
        
        return self.fit_data | { "N_new": X_new.shape[0], "X_new": X_new}



import matplotlib.pyplot
import numpy
import seaborn
import scipy.stats

def parameters_plot(model, include=None, exclude=None, **kwargs):
    if include is None:
        include = model.draws.columns
    if exclude is None:
        exclude = []
    
    columns = [
        x for x in model.draws.columns
        if x in include and x not in exclude]
    
    kwargs.setdefault("estimator", numpy.median)
    kwargs.setdefault("errorbar", ("pi", 90))
    
    ax = seaborn.pointplot(
        model.draws[columns].melt(),
        x="value", y="variable",
        linestyle="none",
        **kwargs)
    ax.set(xlabel=model.outcomes.columns[0], ylabel=None)

def predictive_plot(model, use_prior=False, count=50, alpha=0.2, plot_kwargs={}):
    y = model.prior_predict if use_prior else model.posterior_predict
    subset = numpy.random.randint(0, len(y), count)
    
    for draw in subset:
        seaborn.kdeplot(y.iloc[draw, :], color="C0", alpha=alpha, **plot_kwargs)
    
    seaborn.kdeplot(
        model.outcomes.values.squeeze(), color="k", alpha=1, **plot_kwargs)
    plot_kwargs.get("ax", matplotlib.pyplot.gca()).set(
        xlabel=model.outcomes.columns[0])

class KDEPlot:
    def __init__(
            self, data, ax=None, prob=None, point_estimate=None, color="black",
            alpha=0.3, **kwargs):
        
        self.ax = ax or matplotlib.pyplot.gca()
        
        linewidth = kwargs.pop("lw", kwargs.pop("linewidth", None))
        
        self.kde = scipy.stats.gaussian_kde(data)
        bw = self.kde.scotts_factor() * data.std(ddof=1)
        
        xs = numpy.linspace(data.min()-bw, data.max()+bw, 200)
        self.dist_line = self.ax.plot(
            xs, self.kde(xs), color=color, linewidth=linewidth)
        
        self.dist_area = None
        if prob is not None:
            self.q = numpy.quantile(data, [0.5-prob/2, 0.5+prob/2])
            xs = xs[(xs>=self.q[0]) & (xs<=self.q[1])]
            self.dist_area = self.ax.fill_between(
                xs, self.kde(xs), color=color, alpha=alpha, lw=0)
        
        self.estimate_line = None
        if point_estimate is not None:
            x = point_estimate(data)
            self.estimate_line = self.ax.plot(
                [x, x], [0, *self.kde(x)], color=color, linewidth=linewidth)
        
        self.ax.set(ylim=0, yticks=[])
        self.ax.spines[["top", "left", "right"]].set_visible(False)



import formulaic
import numpy
import pandas

from .predictor_mapper import PredictorMapper

class ModelData:
    def __init__(self, formula, data):
        self.formula = formula
        self.data = data
        
        self.outcomes, self.predictors = zip(
            *[formulaic.model_matrix(f, data) for f in formula])
        self.outcomes = pandas.concat(self.outcomes, axis="columns")
        
        self.predictor_mapper = PredictorMapper(self.predictors, self.outcomes)
        
        mu_y = numpy.mean(self.outcomes.values, axis=0)
        sigma_y = numpy.atleast_1d(numpy.std(self.outcomes.values, axis=0))
        sigma_X = [
            numpy.std(x.filter(regex="^(?!.*Intercept)").values, axis=0)
            for x in self.predictors]
        
        self.fit_data = {
            "R": len(self.formula),
            "N": len(data),
            "K": numpy.squeeze([x.shape[1] for x in self.predictors]),
            "y": numpy.squeeze(self.outcomes.values),
            "X": pandas.concat(self.predictors, axis="columns"),
            
            "mu_alpha": numpy.squeeze(mu_y),
            "sigma_alpha": 2.5*numpy.squeeze(sigma_y),
            "sigma_beta": numpy.concatenate(numpy.atleast_2d(
                [2.5*(sy/sx) for sx, sy in zip(sigma_X, sigma_y)])),
            "lambda_sigma": numpy.squeeze(1/sigma_y),
            "eta_L": 1.0}
    
    def new_predictors(self, data):
        data = data.astype({
            k: v for k, v in self.data.dtypes.items() if k in data.columns})
        predictors = []
        for formula in self.formula:
            predictors.append(
                formulaic.model_matrix(formula.split("~")[1], data))
        predictors = pandas.concat(predictors, axis="columns")
        return predictors
        
    
    def new_data(self, X_new=None):
        if X_new is None:
            X_new = self.fit_data["X"]
        
        return self.fit_data | { "N_new": X_new.shape[0], "X_new": X_new}



import argparse
import os
import re
import shlex
import subprocess
import sys
import tempfile
import textwrap

def compile(stan_file, h_file, include_path=None, prefix=None):
    name = os.path.splitext(os.path.basename(stan_file))[0]
    if prefix:
        name = f"{prefix}{name}"
    
    subprocess.check_call([
        f"{os.environ['CMDSTAN']}/bin/stanc",
        *(["--include-paths", include_path] if include_path else []),
        stan_file, "--name", "model", "--O1",
        "--o", h_file])
    
    with open(h_file) as fd:
        contents = fd.read()
    
    contents = re.sub("model_namespace", name, contents)
    contents = re.sub("using stan_model = [^;]+", "", contents)
    contents = re.sub("stan_model", f"{name}::model", contents)
    contents = re.sub("new_model", f"new_{name}", contents)
    contents = re.sub(
        "get_stan_profile_data", f"get_stan_profile_data_{name}", contents)
    
    with open(h_file, "w") as fd:
        fd.write(contents)

def stan_info(names, optimize=True, use_threads=True, range_checks=False):
    if isinstance(names, str):
        names = [names]
    
    makefile = textwrap.dedent(f"""\
        STAN = $(CMDSTAN)/stan/
        STAN_CPP_OPTIMS = {int(optimize)}
        STAN_THREADS = {int(use_threads)}
        STAN_NO_RANGE_CHECKS = {(int(not range_checks))}

        include $(CMDSTAN)/makefile

        cxxflags:
            @echo $(CPPFLAGS) $(CXXFLAGS)
        ldflags:
            @echo $(LDFLAGS) | sed 's/-Wl,-L,/-L/g'
        libs:
            @echo $(LDLIBS)
    """)
    makefile = re.sub("    ", "\t", makefile)
    with tempfile.TemporaryDirectory() as dir:
        path = os.path.join(dir, "makefile")
        with open(path, "w") as fd:
            fd.write(makefile)
        flags = subprocess.check_output([
            "make", "-I", os.environ["CMDSTAN"], "-f", path, *names])
    return shlex.split(flags.decode())

def main():
    try:
        command = sys.argv[1]
    except IndexError:
        print("Missing command")
        return 1
    
    if command == "compile":
        parser = argparse.ArgumentParser()
        parser.add_argument("stan_file")
        parser.add_argument("h_file")
        parser.add_argument("--include-path", "-I")
        parser.add_argument("--prefix", "-p")
        arguments = parser.parse_args(sys.argv[2:])
        compile(**vars(arguments))
    elif command == "info":
        parser = argparse.ArgumentParser()
        parser.add_argument("names", metavar="name", nargs="+")
        parser.add_argument("--no-optimize", dest="optimize", action="store_false")
        parser.add_argument("--no-use-threads", dest="use_threads", action="store_false")
        parser.add_argument("--range-checks", dest="use_threads", action="store_true")
        arguments = parser.parse_args(sys.argv[2:])
        print(shlex.join(stan_info(**vars(arguments))))
    else:
        print(f"Unknown command: {command}")
        return 1

if __name__ == "__main__":
    sys.exit(main())



import formulaic
import numpy
import pandas

from .predictor_mapper import PredictorMapper

class ModelData:
    def __init__(self, formula, data):
        self.formula = formula
        self.data = data
        
        self.outcomes, self.unmodeled_predictors = formulaic.model_matrix(
            formula[0], data)
        
        group_name, group_formula = formula[1]
        self.modeled_predictors = formulaic.model_matrix(group_formula, data)
        self.modeled_predictors.index = data[group_name]
        
        self.predictor_mapper = PredictorMapper(
            self.unmodeled_predictors, self.modeled_predictors, self.outcomes)
        
        mu_y = numpy.mean(self.outcomes.values)
        sigma_y = numpy.std(self.outcomes.values)
        sigma_X = numpy.std(
            self.unmodeled_predictors.filter(regex="^(?!.*Intercept)").values,
            axis=0)
        
        self.fit_data = {
            "N": len(data),
            "K0": self.unmodeled_predictors.shape[1],
            "K": self.modeled_predictors.shape[1],
            "J": data[group_name].nunique(),
            
            "y": numpy.squeeze(self.outcomes),
            "X0": self.unmodeled_predictors,
            "X": self.modeled_predictors,
            
            "group": 1+data[group_name].cat.codes,
            
            "mu_alpha": mu_y, "sigma_alpha": 2.5*sigma_y,
            "sigma_beta": 2.5*sigma_y/sigma_X,
            
            "lambda_sigma_y": 1/sigma_y,
            "lambda_sigma_Beta": 1/sigma_y,
            
            "eta_L": 1.}
    
    @property
    def predictors(self):
        return (self.unmodeled_predictors, self.modeled_predictors)
    
    def new_data(self, X0_new=None, X_new=None):
        if X0_new is None:
            X0_new = self.fit_data["X0"]
        if X_new is None:
            X_new = self.fit_data["X"]
        
        return self.fit_data | {
            "N_new": X0_new.shape[0], "X0_new": X0_new, "X_new": X_new}



import re

class PredictorMapper:
    """ Map the low-level stan names to the high-level predictor names
    """
    
    def __init__(self, unmodeled_predictors, modeled_predictors, outcomes):
        self._common_names = {
            "alpha": "Intercept",
            "alpha_c": "Intercept_c",
            "sigma": "sigma"}
        
        self._outcomes = {1+i: c for i, c in enumerate(outcomes.columns)}
        
        self._beta = {
            1+index: name
            for index, name in enumerate(
                unmodeled_predictors.filter(regex="^(?!Intercept)").columns)}
        
        self._group_name = modeled_predictors.index.name
        self._groups = modeled_predictors.index.categories
        self._Beta = {
            1+index: name
            for index, name in enumerate(modeled_predictors.columns)}
        
    def __call__(self, x):
        if not isinstance(x, str):
            return [self.__call__(item) for item in x]
        
        match = re.match("([^.]+)\.(\d+)(?:\.(\d+))?", x)
        if match:
            kind, a, b = match.groups()
            if b is not None:
                group, index = int(a), int(b)
            else:
                index = int(a)
        else:
            kind = x
            index = None
            
        if kind in self._common_names:
            return self._common_names[kind]
        elif kind == "beta":
            return self._beta[index]
        elif kind == "Beta":
            coefficient = self._Beta[index]
            group = self._groups[group-1]
            return f"{self._group_name}[{group}]/{coefficient}"
        elif kind == "Sigma_Beta":
            return f"{kind}[{self._Beta[group]}, {self._Beta[index]}]"
        elif kind.endswith("_") and not kind.endswith("__"):
            return f"{kind[:-1]}[{index}]"
        else:
            return x



import glob
import os
import subprocess
import sys
import tempfile

import setuptools
import setuptools.command.build

here = os.path.abspath(os.path.dirname(__file__))

class BuildCMake(setuptools.Command, setuptools.command.build.SubCommand):
    def __init__(self, *args, **kwargs):
        setuptools.Command.__init__(self, *args, **kwargs)
        setuptools.command.build.SubCommand.__init__(self, *args, **kwargs)
        self.build_lib = None
        self.editable_mode = False
        
        self.sources = []
        self.stanc_options = ""
        self.cxxflags = ""
    
    def initialize_options(self):
        pass
        
    def finalize_options(self):
        self.sources = [
            *sorted(glob.glob(f"src/stan/*.stan")),
            *sorted(glob.glob(f"src/stan/*/*.stan")),
            *sorted(glob.glob(f"src/lib/slimp/*")),
            *sorted(glob.glob(f"src/python/*.cpp"))]
        self.set_undefined_options("build_py", ("build_lib", "build_lib"))
    
    def run(self):
        with tempfile.TemporaryDirectory() as build_dir:
            subprocess.check_call(
                [
                    "cmake", f"-DPython_EXECUTABLE={sys.executable}",
                    "-DCMAKE_BUILD_TYPE=Release",
                    "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY="
                        f"{os.path.join(here, self.build_lib, 'slimp')}", 
                    "-S", here, "-B", build_dir])
            
            subprocess.check_call(
                [
                    "cmake", "--build", build_dir,
                    "--target", "libslimp", "--target", "pyslimp",
                    "--config", "Release", "--parallel"])
    
    def get_source_files(self):
        return self.sources

setuptools.command.build.build.sub_commands.append(("build_cmake", None))

long_description = open(os.path.join(here, "README.md")).read()
setuptools.setup(
    name="slimp",
    version="1.0.0",
    
    description="Linear models with Stan and Pandas",
    long_description=long_description,
    long_description_content_type="text/markdown",
    
    url="https://github.com/lamyj/slimp",
    
    author="Julien Lamy",
    author_email="lamy@unistra.fr",
    
    license="MIT",
    
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Framework :: Matplotlib",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering",
    ],
    
    keywords = [
        "statistics", "bayesian", "linear-models", "stan", "pandas",
        "matplotlib"],
    
    cmdclass={"build_cmake": BuildCMake},

    packages=setuptools.find_packages(where="src/python"),
    package_dir={"": "src/python"},
    
    install_requires=[
        "formulaic",
        "numpy",
        "matplotlib",
        "pandas",
        "seaborn"
    ],
)



import pandas
import xarray

def sample_data_as_df(data):
    return pandas.DataFrame(
        data["array"].reshape((data["array"].shape[0], -1), order="A").T,
        columns=data["columns"])

def sample_data_as_xarray(data):
    return xarray.DataArray(
        data["array"],
        dims=["parameter", "chain", "sample"],
        coords={"parameter": data["columns"]})



import pandas

class Samples:
    def __init__(self, samples, predictor_mapper, parameters_columns):
        self.predictor_mapper = predictor_mapper
        
        self.samples = samples
        self.samples["parameter"] = self.predictor_mapper(
            self.samples["parameter"].values)
        
        names = samples["parameter"].values
        diagnostics_names = [x for x in names if x.endswith("_")]
        self.diagnostics = self.samples.sel(parameter=diagnostics_names)
        
        parameters_names = [x for x in names if not x.endswith("_")]
        self.draws = pandas.DataFrame(
            self.samples.sel(parameter=parameters_names)
                .values
                .reshape(len(parameters_names), -1)
                .T,
            columns=predictor_mapper(parameters_names))
        
        self.parameters_columns = parameters_columns



import formulaic
import numpy
import pandas

from . import _slimp, action_parameters, misc, stats
from .samples import Samples

from . import multilevel, multivariate, univariate

class Model:
    def __init__(
            self, formula, data, seed=-1, num_chains=1, sampler_parameters=None,
            **kwargs):
        ModelData = None
        if isinstance(formula, str):
            ModelData = univariate.ModelData
        elif isinstance(formula, list):
            if len(formula) == 2 and isinstance(formula[1], tuple):
                ModelData = multilevel.ModelData
            else:
                ModelData = multivariate.ModelData
        self._model_data = ModelData(formula, data)
        self._model_name = ModelData.__module__.split(".")[1]
        
        if sampler_parameters is None:
            self._sampler_parameters = action_parameters.Sample(
                seed=seed, num_chains=num_chains, **kwargs)
        else:
            self._sampler_parameters = sampler_parameters
        
        self._samples = None
        self._generated_quantities = {}
    
    @property
    def formula(self):
        return (
            self._model_data.formula if len(self._model_data.formula)>1
            else self._model_data.formula[0])
    
    @property
    def data(self):
        return self._model_data.data
    
    @property
    def predictors(self):
        return self._model_data.predictors
    
    @property
    def outcomes(self):
        return self._model_data.outcomes
    
    @property
    def fit_data(self):
        return self._model_data.fit_data
    
    @property
    def sampler_parameters(self):
        return self._sampler_parameters
    
    @property
    def draws(self):
        return self._samples.draws if self._samples is not None else None
    
    @property
    def prior_predict(self):
        if "y_prior" not in self._generated_quantities:
            draws = self._generate_quantities("predict_prior")
            self._generated_quantities["y_prior"] = draws.filter(like="y")
        return self._generated_quantities["y_prior"]
    
    @property
    def posterior_epred(self):
        if "mu_posterior" not in self._generated_quantities:
            draws = self._generate_quantities("predict_posterior")
            self._generated_quantities["mu_posterior"] = draws.filter(like="mu")
            self._generated_quantities["y_posterior"] = draws.filter(like="y")
        return self._generated_quantities["mu_posterior"]
    
    @property
    def posterior_predict(self):
        if "y_posterior" not in self._generated_quantities:
            # Update cached data
            self.posterior_epred
        return self._generated_quantities["y_posterior"]
    
    @property
    def log_likelihood(self):
        if "log_likelihood" not in self._generated_quantities:
            draws = self._generate_quantities("log_likelihood")
            self._generated_quantities["log_likelihood"] = draws.filter(like="log_likelihood")
        return self._generated_quantities["log_likelihood"]
    
    @property
    def hmc_diagnostics(self):
        return stats.hmc_diagnostics(
            self._samples.diagnostics, self._sampler_parameters.hmc.max_depth)
    
    def sample(self, sampler=None):
        if sampler is None:
            sampler = getattr(_slimp, f"{self._model_name}_sampler")
        data = sampler(self._model_data.fit_data, self._sampler_parameters)
        self._samples = Samples(
            misc.sample_data_as_xarray(data),
            self._model_data.predictor_mapper, data["parameters_columns"])
        self._generated_quantities = {}
    
    def summary(self, percentiles=(5, 50, 95)):
        return stats.summary(
            self._samples.samples.sel(
                parameter=["lp__", *self._samples.draws.columns]),
            percentiles)
    
    def predict(self, data):
        predictors = self._model_data.new_predictors(data)
        draws = self._generate_quantities(
            "predict_posterior", predictors.values)
        return draws.filter(like="mu"), draws.filter(like="y")
    
    def _generate_quantities(self, name, *args, **kwargs):
        new_data = self._model_data.new_data(*args, **kwargs)
        
        # NOTE: must only include model parameters
        draws = self._samples.samples.sel(
            parameter=self._samples.predictor_mapper(
                self._samples.parameters_columns))
        data = getattr(_slimp, f"{self._model_name}_{name}")(
            new_data, draws, self._sampler_parameters)
        
        return misc.sample_data_as_df(data)
    
    def __getstate__(self):
        return {
            "formula": self.formula, "data": self.data,
            "sampler_parameters": self._sampler_parameters,
            "model_name": self._model_name,
            **(
                {
                    "samples": self._samples.samples,
                    "parameters_columns": self._samples.parameters_columns}
                if self._samples is not None else {}),
            "generated_quantities": self._generated_quantities
        }
    
    def __setstate__(self, state):
        self.__init__(state["formula"], state["data"])
        self._sampler_parameters = state["sampler_parameters"]
        self._model_name = state["model_name"]
        if "samples" in state:
            self._samples = Samples(
                state["samples"], self._model_data.predictor_mapper,
                state["parameters_columns"])
        self._generated_quantities = state["generated_quantities"]
